@@ -1,15 +1,13 @@
 /*
 -*- linux-c -*-
    drbd_bitmap.c
-   Kernel module for 2.4.x/2.6.x Kernels
+   Kernel module for 2.6.x Kernels
 
-   This file is part of drbd by Philipp Reisner.
+   This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2004, Lars Ellenberg <l.g.e@web.de>.
-	main author.
-
-   Copyright (C) 2004, Philipp Reisner <philipp.reisner@linbit.com>.
-	contributions.
+   Copyright (C) 2004-2007, LINBIT Information Technologies GmbH.
+   Copyright (C) 2004-2007, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2004-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -29,53 +27,12 @@
 #include <linux/bitops.h>
 #include <linux/vmalloc.h>
 #include <linux/string.h> // for memset
+#include <linux/hardirq.h> /* for D_ASSERT(in_interrupt()) */
+
 
 #include <linux/drbd.h>
 #include "drbd_int.h"
 
-/* special handling for ppc64 on 2.4 kernel -- find_next_bit is not exported
- * so we include it here (verbatim, from linux 2.4.21 sources) */
-#if defined(__powerpc64__) && LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
-
-unsigned long find_next_bit(unsigned long *addr, unsigned long size, unsigned long offset)
-{
-        unsigned long *p = addr + (offset >> 6);
-        unsigned long result = offset & ~63UL;
-        unsigned long tmp;
-
-        if (offset >= size)
-                return size;
-        size -= result;
-        offset &= 63UL;
-        if (offset) {
-                tmp = *(p++);
-                tmp &= (~0UL << offset);
-                if (size < 64)
-                        goto found_first;
-                if (tmp)
-                        goto found_middle;
-                size -= 64;
-                result += 64;
-        }
-        while (size & ~63UL) {
-                if ((tmp = *(p++)))
-                        goto found_middle;
-                result += 64;
-                size -= 64;
-        }
-        if (!size)
-                return result;
-        tmp = *p;
-
-found_first:
-        tmp &= (~0UL >> (64 - size));
-        if (tmp == 0UL)        /* Are any bits set? */
-                return result + size; /* Nope. */
-found_middle:
-        return result + __ffs(tmp);
-}
-#endif /* NEED_PPC64_WORKAROUND */
-
 /* OPAQUE outside this file!
  * interface defined in drbd_int.h
  *
@@ -98,14 +55,20 @@ found_middle:
  * or wether I want it to do all the sector<->bit calculation in here.
  */
 
+// warning LGE "verify all spin_lock_irq here, and their call path"
+// warning LGE "and change to irqsave where applicable"
+// warning LGE "so we don't accidentally nest spin_lock_irq()"
 /*
  * NOTE
  *  Access to the *bm is protected by bm_lock.
  *  It is safe to read the other members within the lock.
  *
  *  drbd_bm_set_bit is called from bio_endio callbacks,
- *  so there we need a spin_lock_irqsave.
- *  Everywhere else we need a spin_lock_irq.
+ *  We may be called with irq already disabled,
+ *  so we need spin_lock_irqsave().
+ * FIXME
+ *  for performance reasons, when we _know_ we have irq disabled, we should
+ *  probably introduce some _in_irq variants, so we know to only spin_lock().
  *
  * FIXME
  *  Actually you need to serialize all resize operations.
@@ -123,6 +86,13 @@ found_middle:
 struct drbd_bitmap {
 	unsigned long *bm;
 	spinlock_t bm_lock;
+	/* WARNING unsigned long bm_fo and friends:
+	 * 32bit number of bit offset is just enough for 512 MB bitmap.
+	 * it will blow up if we make the bitmap bigger...
+	 * not that it makes much sense to have a bitmap that large,
+	 * rather change the granularity to 16k or 64k or something.
+	 * (that implies other problems, however...)
+	 */
 	unsigned long bm_fo;        // next offset for drbd_bm_find_next
 	unsigned long bm_set;       // nr of set bits; THINK maybe atomic_t ?
 	unsigned long bm_bits;
@@ -130,8 +100,12 @@ struct drbd_bitmap {
 	sector_t bm_dev_capacity;
 	struct semaphore bm_change; // serializes resize operations
 
+	atomic_t bm_async_io;
+	wait_queue_head_t bm_io_wait;
+
+	unsigned long  bm_flags;
+
 	// { REMOVE
-	unsigned long  bm_flags;     // currently debugging aid only
 	unsigned long  bm_line;
 	char          *bm_file;
 	// }
@@ -140,6 +114,8 @@ struct drbd_bitmap {
 // { REMOVE once we serialize all state changes properly
 #define D_BUG_ON(x)	ERR_IF(x) { dump_stack(); }
 #define BM_LOCKED 0
+#define BM_MD_IO_ERROR (BITS_PER_LONG-1) // 31? 63?
+
 #if 0 // simply disabled for now...
 #define MUST_NOT_BE_LOCKED() do {					\
 	if (test_bit(BM_LOCKED,&b->bm_flags)) {				\
@@ -162,7 +138,7 @@ struct drbd_bitmap {
 #else
 #define MUST_NOT_BE_LOCKED() do {(void)b;} while (0)
 #define MUST_BE_LOCKED() do {(void)b;} while (0)
-#endif 
+#endif
 void __drbd_bm_lock(drbd_dev *mdev, char* file, int line)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
@@ -215,7 +191,12 @@ void drbd_bm_unlock(drbd_dev *mdev)
 #else
 #define BM_PARANOIA_CHECK() do {					\
 	D_ASSERT(b->bm[b->bm_words] == DRBD_MAGIC);			\
-	D_ASSERT(b->bm_dev_capacity == drbd_get_capacity(mdev->this_bdev));	\
+	if (b->bm_dev_capacity != drbd_get_capacity(mdev->this_bdev)) {	\
+		ERR("%s:%d: bm_dev_capacity:%llu drbd_get_capacity:%llu\n", \
+		__FILE__, __LINE__,					\
+		(unsigned long long) b->bm_dev_capacity,		\
+		(unsigned long long) drbd_get_capacity(mdev->this_bdev));\
+	}								\
 } while (0)
 #endif
 // }
@@ -263,13 +244,15 @@ int drbd_bm_init(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	D_BUG_ON(b);
-	b = kmalloc(sizeof(struct drbd_bitmap),GFP_KERNEL);
+	b = kzalloc(sizeof(struct drbd_bitmap),GFP_KERNEL);
 	if (!b)
 		return -ENOMEM;
-	memset(b,0,sizeof(*b));
-	b->bm_lock = SPIN_LOCK_UNLOCKED;
+	spin_lock_init(&b->bm_lock);
 	init_MUTEX(&b->bm_change);
+	init_waitqueue_head(&b->bm_io_wait);
+
 	mdev->bitmap = b;
+
 	return 0;
 }
 
@@ -314,7 +297,7 @@ STATIC int bm_clear_surplus(struct drbd_bitmap * b)
 		cleared += hweight_long(b->bm[w]);
 		b->bm[w++]=0;
 	}
-	
+
 	return cleared;
 }
 
@@ -332,19 +315,39 @@ STATIC void bm_set_surplus(struct drbd_bitmap * b)
 	}
 }
 
-STATIC unsigned long bm_count_bits(struct drbd_bitmap * b)
+STATIC unsigned long bm_count_bits(struct drbd_bitmap * b, int just_read)
 {
 	unsigned long *bm = b->bm;
 	unsigned long *ep = b->bm + b->bm_words;
 	unsigned long bits = 0;
 
 	while ( bm < ep ) {
+		/* on little endian, this is *bm = *bm;
+		 * and should be optimized away by the compiler */
+		if (just_read) *bm = lel_to_cpu(*bm);
 		bits += hweight_long(*bm++);
 	}
 
 	return bits;
 }
 
+void _drbd_bm_recount_bits(drbd_dev *mdev, char* file, int line)
+{
+	struct drbd_bitmap *b = mdev->bitmap;
+	unsigned long flags, bits;
+
+	ERR_IF(!b) return;
+
+	spin_lock_irqsave(&b->bm_lock,flags);
+	bits = bm_count_bits(b,0);
+	if(bits != b->bm_set) {
+		ERR("bm_set was %lu, corrected to %lu. %s:%d\n",
+		    b->bm_set,bits,file,line);
+		b->bm_set = bits;
+	}
+	spin_unlock_irqrestore(&b->bm_lock,flags);
+}
+
 #define BM_SECTORS_PER_BIT (BM_BLOCK_SIZE/512)
 
 /*
@@ -368,6 +371,9 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		down(&b->bm_change);
 	}
 
+	INFO("drbd_bm_resize called with capacity == %llu\n",
+			(unsigned long long)capacity);
+
 	if (capacity == b->bm_dev_capacity)
 		goto out;
 
@@ -383,8 +389,7 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		spin_unlock_irq(&b->bm_lock);
 		goto free_obm;
 	} else {
-		bits  = ALIGN(capacity,BM_SECTORS_PER_BIT)
-		      >> (BM_BLOCK_SIZE_B-9);
+		bits = BM_SECT_TO_BIT(ALIGN(capacity,BM_SECTORS_PER_BIT));
 
 		/* if we would use
 		   words = ALIGN(bits,BITS_PER_LONG) >> LN2_BPL;
@@ -393,7 +398,7 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		*/
 		words = ALIGN(bits,64) >> LN2_BPL;
 
-		D_ASSERT(bits < ((MD_RESERVED_SIZE<<1)-MD_BM_OFFSET)<<12 );
+		D_ASSERT((u64)bits <= (((u64)mdev->bc->md.md_size_sect-MD_BM_OFFSET) << 12));
 
 		if ( words == b->bm_words ) {
 			/* optimize: capacity has changed,
@@ -412,6 +417,7 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 			bytes = (words+1)*sizeof(long);
 			nbm = vmalloc(bytes);
 			if (!nbm) {
+				ERR("bitmap: failed to vmalloc %lu bytes\n",bytes);
 				err = -ENOMEM;
 				goto out;
 			}
@@ -419,6 +425,7 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		spin_lock_irq(&b->bm_lock);
 		obm = b->bm;
 		// brgs. move several MB within spinlock...
+		// FIXME this should go into userspace!
 		if (obm) {
 			bm_set_surplus(b);
 			D_ASSERT(b->bm[b->bm_words] == DRBD_MAGIC);
@@ -426,8 +433,9 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		}
 		growing = words > b->bm_words;
 		if (growing) { // set all newly allocated bits
-			memset( nbm+b->bm_words, -1,
-				(words - b->bm_words) * sizeof(long) );
+			// start at -1, just to be sure.
+			memset( nbm + (b->bm_words?:1)-1 , 0xff,
+				(words - ((b->bm_words?:1)-1)) * sizeof(long) );
 			b->bm_set  += bits - b->bm_bits;
 		}
 		nbm[words] = DRBD_MAGIC;
@@ -436,7 +444,7 @@ int drbd_bm_resize(drbd_dev *mdev, sector_t capacity)
 		b->bm_words = words;
 		b->bm_dev_capacity = capacity;
 		bm_clear_surplus(b);
-		if( !growing ) b->bm_set = bm_count_bits(b);
+		if( !growing ) b->bm_set = bm_count_bits(b,0);
 		bm_end_info(mdev, __FUNCTION__ );
 		spin_unlock_irq(&b->bm_lock);
 		INFO("resync bitmap: bits=%lu words=%lu\n",bits,words);
@@ -497,6 +505,7 @@ void drbd_bm_merge_lel( drbd_dev *mdev, size_t offset, size_t number,
 	unsigned long word, bits;
 	size_t n = number;
 
+	if (number == 0) return;
 	ERR_IF(!b) return;
 	ERR_IF(!b->bm) return;
 	D_BUG_ON(offset        >= b->bm_words);
@@ -537,6 +546,7 @@ void drbd_bm_set_lel( drbd_dev *mdev, size_t offset, size_t number,
 	unsigned long word, bits;
 	size_t n = number;
 
+	if (number == 0) return;
 	ERR_IF(!b) return;
 	ERR_IF(!b->bm) return;
 	D_BUG_ON(offset        >= b->bm_words);
@@ -575,6 +585,7 @@ void drbd_bm_get_lel( drbd_dev *mdev, size_t offset, size_t number,
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long *bm;
 
+	if (number == 0) return;
 	ERR_IF(!b) return;
 	ERR_IF(!b->bm) return;
 	if ( (offset        >= b->bm_words) ||
@@ -609,12 +620,64 @@ void drbd_bm_set_all(drbd_dev *mdev)
 
 	spin_lock_irq(&b->bm_lock);
 	BM_PARANOIA_CHECK();
-	memset(b->bm,-1,b->bm_words*sizeof(long));
+	memset(b->bm,0xff,b->bm_words*sizeof(long));
 	bm_clear_surplus(b);
 	b->bm_set = b->bm_bits;
 	spin_unlock_irq(&b->bm_lock);
 }
 
+int drbd_bm_async_io_complete(struct bio *bio, unsigned int bytes_done, int error)
+{
+	struct drbd_bitmap *b = bio->bi_private;
+
+	if (bio->bi_size)
+		return 1;
+
+	if (error) {
+		/* doh. what now?
+		 * for now, set all bits, and flag MD_IO_ERROR
+		 */
+		/* FIXME kmap_atomic memset etc. pp. */
+		__set_bit(BM_MD_IO_ERROR,&b->bm_flags);
+	}
+	if (atomic_dec_and_test(&b->bm_async_io))
+		wake_up(&b->bm_io_wait);
+
+	bio_put(bio);
+
+	return 0;
+}
+
+STATIC void drbd_bm_page_io_async(drbd_dev *mdev, struct drbd_bitmap *b, int page_nr, int rw)
+{
+	/* we are process context. we always get a bio */
+	/* THINK: do we need GFP_NOIO here? */
+	struct bio *bio = bio_alloc(GFP_KERNEL, 1);
+	struct page *page = vmalloc_to_page((char*)(b->bm) + (PAGE_SIZE*page_nr));
+	unsigned int len;
+	sector_t on_disk_sector = mdev->bc->md.md_offset + mdev->bc->md.bm_offset;
+	on_disk_sector += ((sector_t)page_nr) << (PAGE_SHIFT-9);
+
+	/* this might happen with very small flexible external meta data device */
+	len = min_t(unsigned int, PAGE_SIZE,
+		(drbd_md_last_sector(mdev->bc) - on_disk_sector + 1)<<9);
+
+	D_DUMPLU(on_disk_sector);
+	D_DUMPI(len);
+
+	bio->bi_bdev = mdev->bc->md_bdev;
+	bio->bi_sector = on_disk_sector;
+	bio_add_page(bio, page, len, 0);
+	bio->bi_private = b;
+	bio->bi_end_io = drbd_bm_async_io_complete;
+
+	if (FAULT_ACTIVE(mdev, (rw&WRITE)?DRBD_FAULT_MD_WR:DRBD_FAULT_MD_RD)) {
+		bio->bi_rw |= rw;
+		bio_endio(bio,bio->bi_size,-EIO);
+	}
+	else
+		submit_bio(rw, bio);
+}
 /* read one sector of the on disk bitmap into memory.
  * on disk bitmap is little endian.
  * @enr is _sector_ offset from start of on disk bitmap (aka bm-extent nr).
@@ -622,18 +685,18 @@ void drbd_bm_set_all(drbd_dev *mdev)
  */
 int drbd_bm_read_sect(drbd_dev *mdev,unsigned long enr)
 {
-	sector_t on_disk_sector = enr + drbd_md_ss(mdev) + MD_BM_OFFSET;
+	sector_t on_disk_sector = mdev->bc->md.md_offset + mdev->bc->md.bm_offset + enr;
 	int bm_words, num_words, offset, err  = 0;
 
 	// MUST_BE_LOCKED(); not neccessarily global ...
 
 	down(&mdev->md_io_mutex);
-	if(drbd_md_sync_page_io(mdev,on_disk_sector,READ)) {
+	if(drbd_md_sync_page_io(mdev,mdev->bc,on_disk_sector,READ)) {
 		bm_words  = drbd_bm_words(mdev);
 		offset    = S2W(enr);	// word offset into bitmap
 		num_words = min(S2W(1), bm_words - offset);
 #if DUMP_MD >= 3
-	INFO("read_sect: sector=%lu offset=%u num_words=%u\n",
+	INFO("read_sect: sector=%lus offset=%u num_words=%u\n",
 			enr, offset, num_words);
 #endif
 		drbd_bm_set_lel( mdev, offset, num_words,
@@ -642,10 +705,10 @@ int drbd_bm_read_sect(drbd_dev *mdev,unsigned long enr)
 		int i;
 		err = -EIO;
 		ERR( "IO ERROR reading bitmap sector %lu "
-		     "(meta-disk sector %lu)\n",
-		     enr, (unsigned long)on_disk_sector );
-		drbd_chk_io_error(mdev, 1);
-		drbd_io_error(mdev);
+		     "(meta-disk sector %llu)\n",
+		     enr, (unsigned long long)on_disk_sector );
+		drbd_chk_io_error(mdev, 1, TRUE);
+		drbd_io_error(mdev, TRUE);
 		for (i = 0; i < AL_EXT_PER_BM_SECT; i++)
 			drbd_bm_ALe_set_all(mdev,enr*AL_EXT_PER_BM_SECT+i);
 	}
@@ -655,26 +718,122 @@ int drbd_bm_read_sect(drbd_dev *mdev,unsigned long enr)
 
 /**
  * drbd_bm_read: Read the whole bitmap from its on disk location.
+ *
+ * currently only called from "drbd_ioctl_set_disk"
+ * FIXME need to be able to return an error!!
+ *
  */
-void drbd_bm_read(struct Drbd_Conf *mdev)
+# if defined(__LITTLE_ENDIAN)
+	/* nothing to do, on disk == in memory */
+# define bm_cpu_to_lel(x) ((void)0)
+# else
+void bm_cpu_to_lel(struct drbd_bitmap *b)
+{
+	/* need to cpu_to_lel all the pages ...
+	 * this may be optimized by using
+	 * cpu_to_lel(-1) == -1 and cpu_to_lel(0) == 0;
+	 * the following is still not optimal, but better than nothing */
+	const unsigned long *end = b->bm+b->bm_words;
+	unsigned long *bm;
+	if (b->bm_set == 0) {
+		/* no page at all; avoid swap if all is 0 */
+		return;
+	} else if (b->bm_set == b->bm_bits) {
+		/* only the last words */
+		bm = end-2;
+	} else {
+		/* all pages */
+		bm = b->bm;
+	}
+	for (; bm < end; bm++) {
+		*bm = cpu_to_lel(*bm);
+	}
+}
+# endif
+/* lel_to_cpu == cpu_to_lel */
+# define bm_lel_to_cpu(x) bm_cpu_to_lel(x)
+
+STATIC int drbd_bm_rw(struct Drbd_Conf *mdev, int rw)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
-	sector_t sector;
-	int bm_words, num_sectors;
+	/* sector_t sector; */
+	int bm_words, num_pages, i;
+	unsigned long now;
 	char ppb[10];
+	int err = 0;
 
 	MUST_BE_LOCKED();
 
 	bm_words    = drbd_bm_words(mdev);
-	num_sectors = (bm_words*sizeof(long) + 511) >> 9;
+	num_pages = (bm_words*sizeof(long) + PAGE_SIZE-1) >> PAGE_SHIFT;
 
-	for (sector = 0; sector < num_sectors; sector++) {
-		// FIXME do something on io error here?
-		drbd_bm_read_sect(mdev,sector);
+	/* OK, I manipulate the bitmap low level,
+	 * and I expect to be the exclusive user.
+	 * If not, I am really in a bad mood...
+	 * to catch such bugs early, make all people who want to access the
+	 * bitmap while I read/write it dereference a NULL pointer :->
+	 */
+	mdev->bitmap = NULL;
+
+	if(rw == WRITE)	bm_cpu_to_lel(b);
+
+	now = jiffies;
+	atomic_set(&b->bm_async_io, num_pages);
+	__clear_bit(BM_MD_IO_ERROR,&b->bm_flags);
+
+	for (i = 0; i < num_pages; i++) {
+		/* let the layers below us try to merge these bios... */
+		drbd_bm_page_io_async(mdev,b,i,rw);
 	}
 
+	drbd_blk_run_queue(bdev_get_queue(mdev->bc->md_bdev));
+	wait_event(b->bm_io_wait, atomic_read(&b->bm_async_io) == 0);
+	INFO("%s of bitmap took %lu jiffies\n",
+	     rw == READ ? "reading" : "writing", jiffies - now);
+
+	if (test_bit(BM_MD_IO_ERROR,&b->bm_flags)) {
+		ALERT("we had at least one MD IO ERROR during bitmap IO\n");
+		drbd_chk_io_error(mdev, 1, TRUE);
+		drbd_io_error(mdev, TRUE);
+		err = -EIO;
+	}
+
+	now = jiffies;
+	if(rw == WRITE) {
+		bm_lel_to_cpu(b);
+	} else /* rw == READ */ {
+		/* just read, if neccessary adjust endianness */
+		b->bm_set = bm_count_bits(b, 1);
+		INFO("recounting of set bits took additional %lu jiffies\n",
+		     jiffies - now);
+	}
+
+	/* ok, done,
+	 * now it is visible again
+	 */
+
+	mdev->bitmap = b;
+
 	INFO("%s marked out-of-sync by on disk bit-map.\n",
 	     ppsize(ppb,drbd_bm_total_weight(mdev) << (BM_BLOCK_SIZE_B-10)) );
+
+	return err;
+}
+
+int drbd_bm_read(struct Drbd_Conf *mdev)
+{
+	struct drbd_bitmap *b = mdev->bitmap;
+	int err=0;
+
+	if (b->bm) {
+	    // bitmap size > 0
+	    err = drbd_bm_rw(mdev, READ);
+
+	    if (err == 0)
+		b->bm[b->bm_words] = DRBD_MAGIC;
+	}
+
+	return err;
 }
 
 /**
@@ -686,7 +845,7 @@ void drbd_bm_read(struct Drbd_Conf *mdev)
  */
 int drbd_bm_write_sect(struct Drbd_Conf *mdev,unsigned long enr)
 {
-	sector_t on_disk_sector = enr + drbd_md_ss(mdev) + MD_BM_OFFSET;
+	sector_t on_disk_sector = enr + mdev->bc->md.md_offset + mdev->bc->md.bm_offset;
 	int bm_words, num_words, offset, err  = 0;
 
 	// MUST_BE_LOCKED(); not neccessarily global...
@@ -704,14 +863,14 @@ int drbd_bm_write_sect(struct Drbd_Conf *mdev,unsigned long enr)
 	}
 	drbd_bm_get_lel( mdev, offset, num_words,
 			 page_address(mdev->md_io_page) );
-	if (!drbd_md_sync_page_io(mdev,on_disk_sector,WRITE)) {
+	if (!drbd_md_sync_page_io(mdev,mdev->bc,on_disk_sector,WRITE)) {
 		int i;
 		err = -EIO;
 		ERR( "IO ERROR writing bitmap sector %lu "
-		     "(meta-disk sector %lu)\n",
-		     enr, (unsigned long)on_disk_sector );
-		drbd_chk_io_error(mdev, 1);
-		drbd_io_error(mdev);
+		     "(meta-disk sector %llus)\n",
+		     enr, (unsigned long long)on_disk_sector );
+		drbd_chk_io_error(mdev, 1, TRUE);
+		drbd_io_error(mdev, TRUE);
 		for (i = 0; i < AL_EXT_PER_BM_SECT; i++)
 			drbd_bm_ALe_set_all(mdev,enr*AL_EXT_PER_BM_SECT+i);
 	}
@@ -723,30 +882,21 @@ int drbd_bm_write_sect(struct Drbd_Conf *mdev,unsigned long enr)
 /**
  * drbd_bm_write: Write the whole bitmap to its on disk location.
  */
-void drbd_bm_write(struct Drbd_Conf *mdev)
+int drbd_bm_write(struct Drbd_Conf *mdev)
 {
-	struct drbd_bitmap *b = mdev->bitmap;
-	sector_t sector;
-	int bm_words, num_sectors;
-
-	MUST_BE_LOCKED();
-
-	bm_words    = drbd_bm_words(mdev);
-	num_sectors = (bm_words*sizeof(long) + 511) >> 9;
-
-	for (sector = 0; sector < num_sectors; sector++) {
-		// FIXME do something on io error here?
-		drbd_bm_write_sect(mdev,sector);
-	}
+	int err = drbd_bm_rw(mdev, WRITE);
 
 	INFO("%lu KB now marked out-of-sync by on disk bit-map.\n",
 	      drbd_bm_total_weight(mdev) << (BM_BLOCK_SIZE_B-10) );
+
+	return err;
 }
 
 /* clear all bits in the bitmap */
 void drbd_bm_clear_all(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
+
 	ERR_IF(!b) return;
 	ERR_IF(!b->bm) return;
 
@@ -762,6 +912,7 @@ void drbd_bm_clear_all(drbd_dev *mdev)
 void drbd_bm_reset_find(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
+
 	ERR_IF(!b) return;
 
 	MUST_BE_LOCKED();
@@ -782,6 +933,7 @@ unsigned long drbd_bm_find_next(drbd_dev *mdev)
 {
 	struct drbd_bitmap *b = mdev->bitmap;
 	unsigned long i = -1UL;
+
 	ERR_IF(!b) return i;
 	ERR_IF(!b->bm) return i;
 
@@ -802,6 +954,19 @@ unsigned long drbd_bm_find_next(drbd_dev *mdev)
 	return i;
 }
 
+void drbd_bm_set_find(drbd_dev *mdev, unsigned long i)
+{
+	struct drbd_bitmap *b = mdev->bitmap;
+
+	spin_lock_irq(&b->bm_lock);
+	BM_PARANOIA_CHECK();
+
+	b->bm_fo = min_t(unsigned long, i, b->bm_bits);
+
+	spin_unlock_irq(&b->bm_lock);
+}
+
+
 int drbd_bm_rs_done(drbd_dev *mdev)
 {
 	return mdev->bitmap->bm_fo == 0;
@@ -827,7 +992,7 @@ int drbd_bm_set_bit(drbd_dev *mdev, const unsigned long bitnr)
 	                 !(test_bit(DISKLESS,&mdev->flags) ||
 	                   test_bit(PARTNER_DISKLESS,&mdev->flags)) );
 	if (strange_state)
-		ERR("%s in drbd_bm_set_bit\n", cstate_to_name(mdev->cstate));
+		ERR("%s in drbd_bm_set_bit\n", conns_to_name(mdev->cstate));
 */
 
 	spin_lock_irq(&b->bm_lock);
@@ -844,6 +1009,39 @@ int drbd_bm_set_bit(drbd_dev *mdev, const unsigned long bitnr)
 	return i;
 }
 
+/* returns number of bits actually changed (0->1)
+ * wants bitnr, not sector */
+int drbd_bm_set_bits_in_irq(drbd_dev *mdev, const unsigned long s, const unsigned long e)
+{
+	struct drbd_bitmap *b = mdev->bitmap;
+	unsigned long bitnr;
+	int c = 0;
+	ERR_IF(!b) return 1;
+	ERR_IF(!b->bm) return 1;
+
+#if 0
+	/* hm. I assumed that, when inside of lock_irq/unlock_irq,
+	 * in_interrupt() would be true ?
+	 * how else can I assert that this called with irq disabled without using
+	 * spin_lock_irqsave? */
+	D_BUG_ON(!in_interrupt()); /* called within spin_lock_irq(&mdev->req_lock) */
+#endif
+
+	spin_lock(&b->bm_lock);
+	BM_PARANOIA_CHECK();
+	MUST_NOT_BE_LOCKED();
+	for (bitnr = s; bitnr <=e; bitnr++) {
+		ERR_IF (bitnr >= b->bm_bits) {
+			ERR("bitnr=%lu bm_bits=%lu\n",bitnr, b->bm_bits);
+		} else {
+			c += (0 == __test_and_set_bit(bitnr, b->bm));
+		}
+	}
+	b->bm_set += c;
+	spin_unlock(&b->bm_lock);
+	return c;
+}
+
 /* returns previous bit state
  * wants bitnr, NOT sector.
  */
@@ -872,7 +1070,7 @@ int drbd_bm_clear_bit(drbd_dev *mdev, const unsigned long bitnr)
 	 * strange_state blubber is already in place there ...
 	if (i && mdev->cstate <= Connected)
 		ERR("drbd_bm_clear_bit: cleared a bitnr=%lu while %s\n",
-				bitnr, cstate_to_name(mdev->cstate));
+				bitnr, conns_to_name(mdev->cstate));
 	 */
 
 	return i;
@@ -881,6 +1079,9 @@ int drbd_bm_clear_bit(drbd_dev *mdev, const unsigned long bitnr)
 /* returns bit state
  * wants bitnr, NOT sector.
  * inherently racy... area needs to be locked by means of {al,rs}_lru
+ *  1 ... bit set
+ *  0 ... bit not set
+ * -1 ... first out of bounds access, stop testing for bits!
  */
 int drbd_bm_test_bit(drbd_dev *mdev, const unsigned long bitnr)
 {
@@ -891,12 +1092,15 @@ int drbd_bm_test_bit(drbd_dev *mdev, const unsigned long bitnr)
 
 	spin_lock_irq(&b->bm_lock);
 	BM_PARANOIA_CHECK();
-	ERR_IF (bitnr >= b->bm_bits) {
-		ERR("bitnr=%lu bm_bits=%lu\n",bitnr, b->bm_bits);
+	if (bitnr < b->bm_bits) {
+		i = test_bit(bitnr, b->bm) ? 1 : 0;
+	} else if (bitnr == b->bm_bits) {
+		i = -1;
+	} else /* (bitnr > b->bm_bits) */ {
+		ERR("bitnr=%lu > bm_bits=%lu\n",bitnr, b->bm_bits);
 		i = 0;
-	} else {
-		i = test_bit(bitnr, b->bm);
 	}
+
 	spin_unlock_irq(&b->bm_lock);
 	return i;
 }
