@@ -1,15 +1,13 @@
 /*
 -*- linux-c -*-
    drbd_proc.c
-   Kernel module for 2.4.x/2.6.x Kernels
+   Kernel module for 2.6.x Kernels
 
-   This file is part of drbd by Philipp Reisner.
+   This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 1999-2004, Philipp Reisner <philipp.reisner@linbit.com>.
-	main author.
-
-   Copyright (C) 2002-2004, Lars Ellenberg <l.g.e@web.de>.
-	main contributor.
+   Copyright (C) 2001-2007, LINBIT Information Technologies GmbH.
+   Copyright (C) 1999-2007, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2002-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -27,7 +25,7 @@
 
  */
 
-#include <linux/config.h>
+#include <linux/autoconf.h>
 #include <linux/module.h>
 
 #include <asm/uaccess.h>
@@ -38,9 +36,9 @@
 #include <linux/seq_file.h>
 #include <linux/drbd.h>
 #include "drbd_int.h"
+#include "lru_cache.h" /* for lc_sprintf_stats */
 
 STATIC int drbd_proc_open(struct inode *inode, struct file *file);
-STATIC int drbd_seq_show(struct seq_file *seq, void *v);
 
 
 struct proc_dir_entry *drbd_proc;
@@ -49,31 +47,7 @@ struct file_operations drbd_proc_fops = {
 	.open		= drbd_proc_open,
 	.read		= seq_read,
 	.llseek		= seq_lseek,
-	.release	= seq_release,
-};
-
-// We ommit single_open and single_release, since that is only available
-// after 2.4.23
-static void *single_start(struct seq_file *p, loff_t *pos)
-{
-	return NULL + (*pos == 0);
-}
-
-static void *single_next(struct seq_file *p, void *v, loff_t *pos)
-{
-	++*pos;
-	return NULL;
-}
-
-static void single_stop(struct seq_file *p, void *v)
-{
-}
-
-struct seq_operations drbd_proc_seq_ops = {
-	.start		= single_start,
-	.next		= single_next,
-	.stop		= single_stop,
-	.show		= drbd_seq_show,
+	.release	= single_release,
 };
 
 /*lge
@@ -100,15 +74,15 @@ STATIC void drbd_syncer_progress(struct Drbd_Conf* mdev, struct seq_file *seq)
 	 * units of BM_BLOCK_SIZE.
 	 * for the percentage, we don't care. */
 
-	rs_left = drbd_bm_total_weight(mdev);
+	rs_left = drbd_bm_total_weight(mdev) - mdev->rs_failed;
 	/* >> 10 to prevent overflow,
 	 * +1 to prevent division by zero */
 	if (rs_left > mdev->rs_total) {
 		/* doh. logic bug somewhere.
 		 * for now, just try to prevent in-kernel buffer overflow.
 		 */
-		ERR("logic bug? rs_left=%lu > rs_total=%lu\n",
-				rs_left, mdev->rs_total);
+		ERR("logic bug? rs_left=%lu > rs_total=%lu (rs_failed %lu)\n",
+				rs_left, mdev->rs_total, mdev->rs_failed);
 		res = 1000;
 	} else {
 		res = (rs_left >> 10)*1000/((mdev->rs_total >> 10) + 1);
@@ -184,49 +158,23 @@ STATIC void drbd_syncer_progress(struct Drbd_Conf* mdev, struct seq_file *seq)
 	seq_printf(seq," K/sec\n");
 }
 
-const char* cstate_to_name(Drbd_CState s) {
-	static const char *cstate_names[] = {
-		[Unconfigured]   = "Unconfigured",
-		[StandAlone]     = "StandAlone",
-		[Unconnected]    = "Unconnected",
-		[Timeout]        = "Timeout",
-		[BrokenPipe]     = "BrokenPipe",
-		[NetworkFailure] = "NetworkFailure",
-		[WFConnection]   = "WFConnection",
-		[WFReportParams] = "WFReportParams",
-		[Connected]      = "Connected",
-		[SkippedSyncS]   = "SkippedSyncS",
-		[SkippedSyncT]   = "SkippedSyncT",
-		[WFBitMapS]      = "WFBitMapS",
-		[WFBitMapT]      = "WFBitMapT",
-		[SyncSource]     = "SyncSource",
-		[SyncTarget]     = "SyncTarget",
-		[PausedSyncS]    = "PausedSyncS",
-		[PausedSyncT]    = "PausedSyncT",
-	};
+#if 0
+STATIC void resync_dump_detail(struct seq_file *seq, struct lc_element * e)
+{
+	struct bm_extent *bme = (struct bm_extent *)e;
 
-	return s < Unconfigured ? "TO_SMALL" :
-	       s > PausedSyncT  ? "TO_LARGE"
-		                : cstate_names[s];
+	seq_printf(seq,"%5d %s %s\n",bme->rs_left,
+		   bme->flags & BME_NO_WRITES ? "NO_WRITES" : "---------",
+		   bme->flags & BME_LOCKED ? "LOCKED" : "------"
+		   );
 }
-
-const char* nodestate_to_name(Drbd_State s) {
-	static const char *state_names[] = {
-		[Primary]   = "Primary",
-		[Secondary] = "Secondary",
-		[Unknown]   = "Unknown"
-	};
-
-	return s < Unknown    ? "TO_SMALL" :
-	       s > Secondary  ? "TO_LARGE"
-		              : state_names[s];
-}
-
+#endif
 
 STATIC int drbd_seq_show(struct seq_file *seq, void *v)
 {
-	int i;
+	int i,hole=0;
 	const char *sn;
+	drbd_dev *mdev;
 
 	seq_printf(seq, "version: " REL_VERSION " (api:%d/proto:%d)\n%s\n",
 		    API_VERSION,PRO_VERSION, drbd_buildtag());
@@ -245,42 +193,67 @@ STATIC int drbd_seq_show(struct seq_file *seq, void *v)
 	*/
 
 	for (i = 0; i < minor_count; i++) {
-		sn = cstate_to_name(drbd_conf[i].cstate);
-		if(drbd_conf[i].cstate == Connected) {
-			if(test_bit(DISKLESS,&drbd_conf[i].flags))
-				sn = "DiskLessClient";
-			if(test_bit(PARTNER_DISKLESS,&drbd_conf[i].flags))
-				sn = "ServerForDLess";
+		mdev = minor_to_mdev(i);
+		if(!mdev) {
+			hole=1;
+			continue;
 		}
-		if ( drbd_conf[i].cstate == Unconfigured )
+		if( hole ) {
+			hole=0;
+			seq_printf( seq, "\n");
+		}
+
+		sn = conns_to_name(mdev->state.conn);
+
+		if ( mdev->state.conn == StandAlone &&
+		     mdev->state.disk == Diskless) {
 			seq_printf( seq, "%2d: cs:Unconfigured\n", i);
-		else
+		} else {
 			seq_printf( seq,
-			   "%2d: cs:%s st:%s/%s ld:%s\n"
+			   "%2d: cs:%s st:%s/%s ds:%s/%s %c %c%c%c%c\n"
 			   "    ns:%u nr:%u dw:%u dr:%u al:%u bm:%u "
 			   "lo:%d pe:%d ua:%d ap:%d\n",
 			   i, sn,
-			   nodestate_to_name(drbd_conf[i].state),
-			   nodestate_to_name(drbd_conf[i].o_state),
-			   (drbd_conf[i].gen_cnt[Flags]
-			    & MDF_Consistent) ? "Consistent" : "Inconsistent",
-			// FIXME partner consistent?
-			   drbd_conf[i].send_cnt/2,
-			   drbd_conf[i].recv_cnt/2,
-			   drbd_conf[i].writ_cnt/2,
-			   drbd_conf[i].read_cnt/2,
-			   drbd_conf[i].al_writ_cnt,
-			   drbd_conf[i].bm_writ_cnt,
-			   atomic_read(&drbd_conf[i].local_cnt),
-			   atomic_read(&drbd_conf[i].ap_pending_cnt) +
-			   atomic_read(&drbd_conf[i].rs_pending_cnt),
-			   atomic_read(&drbd_conf[i].unacked_cnt),
-			   atomic_read(&drbd_conf[i].ap_bio_cnt)
+			   roles_to_name(mdev->state.role),
+			   roles_to_name(mdev->state.peer),
+			   disks_to_name(mdev->state.disk),
+			   disks_to_name(mdev->state.pdsk),
+			   (mdev->net_conf == NULL ? ' ' :
+			    (mdev->net_conf->wire_protocol - DRBD_PROT_A+'A')),
+			   mdev->state.susp ? 's' : 'r',
+			   mdev->state.aftr_isp ? 'a' : '-',
+			   mdev->state.peer_isp ? 'p' : '-',
+			   mdev->state.user_isp ? 'u' : '-',
+			   mdev->send_cnt/2,
+			   mdev->recv_cnt/2,
+			   mdev->writ_cnt/2,
+			   mdev->read_cnt/2,
+			   mdev->al_writ_cnt,
+			   mdev->bm_writ_cnt,
+			   atomic_read(&mdev->local_cnt),
+			   atomic_read(&mdev->ap_pending_cnt) +
+			   atomic_read(&mdev->rs_pending_cnt),
+			   atomic_read(&mdev->unacked_cnt),
+			   atomic_read(&mdev->ap_bio_cnt)
 			);
+		}
+		if ( mdev->state.conn == SyncSource ||
+		     mdev->state.conn == SyncTarget ) {
+			drbd_syncer_progress(mdev,seq);
+		}
+		if(mdev->resync) {
+			lc_printf_stats(seq,mdev->resync);
+		}
+		if(mdev->act_log) {
+			lc_printf_stats(seq,mdev->act_log);
+		}
+#if 0
+		if(mdev->resync) {
+			lc_dump(mdev->resync,seq,"rs_left",
+				resync_dump_detail);
+		}
+#endif
 
-		if ( drbd_conf[i].cstate == SyncSource ||
-		     drbd_conf[i].cstate == SyncTarget )
-			drbd_syncer_progress(drbd_conf+i,seq);
 	}
 
 	return 0;
@@ -288,7 +261,7 @@ STATIC int drbd_seq_show(struct seq_file *seq, void *v)
 
 STATIC int drbd_proc_open(struct inode *inode, struct file *file)
 {
-	return seq_open(file, &drbd_proc_seq_ops);
+	return single_open(file, drbd_seq_show, PDE(inode)->data);
 }
 
 /* PROC FS stuff end */
