@@ -1,13 +1,13 @@
 /*
 -*- linux-c -*-
    lru_cache.c
-   Kernel module for 2.4.x/2.6.x Kernels
+   Kernel module for 2.6.x Kernels
 
-   This file is part of drbd by Philipp Reisner.
+   This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2003-2004, Philipp Reisner <philipp.reisner@linbit.com>.
-   Copyright (C) 2003-2004, Lars Ellenberg <l.g.e@web.de>.
-        authors.
+   Copyright (C) 2003-2007, LINBIT Information Technologies GmbH.
+   Copyright (C) 2003-2007, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2003-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,10 +24,11 @@
    the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
  */
-#include <linux/compiler.h> // for likely()
+
 #include <linux/bitops.h>
 #include <linux/vmalloc.h>
 #include <linux/string.h> // for memset
+#include <linux/seq_file.h> // for seq_printf
 #include "lru_cache.h"
 
 #define STATIC static
@@ -42,8 +43,8 @@
  * struct lru_cache, and the hash table slots.
  * returns pointer to a newly initialized lru_cache object with said parameters.
  */
-struct lru_cache* lc_alloc(unsigned int e_count, size_t e_size,
-			   void *private_p)
+struct lru_cache* lc_alloc(const char *name, unsigned int e_count,
+			   size_t e_size, void *private_p)
 {
 	unsigned long bytes;
 	struct lru_cache   *lc;
@@ -65,6 +66,7 @@ struct lru_cache* lc_alloc(unsigned int e_count, size_t e_size,
 		lc->nr_elements      = e_count;
 		lc->new_number	     = -1;
 		lc->lc_private       = private_p;
+		lc->name             = name;
 		for(i=0;i<e_count;i++) {
 			e = lc_entry(lc,i);
 			e->lc_number = LC_FREE;
@@ -84,6 +86,20 @@ void lc_free(struct lru_cache* lc)
 	vfree(lc);
 }
 
+size_t	lc_printf_stats(struct seq_file *seq, struct lru_cache* lc)
+{
+	/* NOTE:
+	 * total calls to lc_get are
+	 * starving + hits + misses
+	 * misses include "dirty" count (update from an other thread in progress)
+	 * and "changed", when this in fact lead to an successful update of the cache.
+	 */
+	return seq_printf(seq,"\t%s: used:%u/%u "
+		"hits:%lu misses:%lu starving:%lu dirty:%lu changed:%lu\n",
+		lc->name, lc->used, lc->nr_elements,
+		lc->hits, lc->misses, lc->starving, lc->dirty, lc->changed);
+}
+
 static unsigned int lc_hash_fn(struct lru_cache* lc, unsigned int enr)
 {
 	return enr % lc->nr_elements;
@@ -137,7 +153,7 @@ void lc_del(struct lru_cache* lc, struct lc_element *e)
 	PARANOIA_ENTRY();
 	BUG_ON(e->refcnt);
 	list_del(&e->list);
-	hlist_del(&e->colision);
+	hlist_del_init(&e->colision);
 	e->lc_number = LC_FREE;
 	e->refcnt = 0;
 	list_add(&e->list,&lc->free);
@@ -199,15 +215,21 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	BUG_ON(!lc->nr_elements);
 
 	PARANOIA_ENTRY();
-	if ( lc->flags & LC_STARVING ) RETURN(NULL);
+	if ( lc->flags & LC_STARVING ) {
+		++lc->starving;
+		RETURN(NULL);
+	}
 
 	e = lc_find(lc, enr);
 	if (e) {
-		++e->refcnt;
+		++lc->hits;
+		if( e->refcnt++ == 0) lc->used++;
 		list_move(&e->list,&lc->in_use); // Not evictable...
 		RETURN(e);
 	}
 
+	++lc->misses;
+
 	/* In case there is nothing available and we can not kick out
 	 * the LRU element, we have to wait ...
 	 */
@@ -220,13 +242,17 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	 * which then is replaced.
 	 * we need to update the cache; serialize on lc->flags & LC_DIRTY
 	 */
-	if (test_and_set_bit(__LC_DIRTY,&lc->flags)) RETURN(NULL);
+	if (test_and_set_bit(__LC_DIRTY,&lc->flags)) {
+		++lc->dirty;
+		RETURN(NULL);
+	}
 
 	e = lc_get_unused_element(lc);
 	BUG_ON(!e);
 
 	clear_bit(__LC_STARVING,&lc->flags);
 	BUG_ON(++e->refcnt != 1);
+	lc->used++;
 
 	lc->changing_element = e;
 	lc->new_number = enr;
@@ -234,10 +260,37 @@ struct lc_element* lc_get(struct lru_cache* lc, unsigned int enr)
 	RETURN(e);
 }
 
+/* similar to lc_get,
+ * but only gets a new reference on an existing element.
+ * you either get the requested element, or NULL.
+ */
+struct lc_element* lc_try_get(struct lru_cache* lc, unsigned int enr)
+{
+	struct lc_element *e;
+
+	BUG_ON(!lc);
+	BUG_ON(!lc->nr_elements);
+
+	PARANOIA_ENTRY();
+	if ( lc->flags & LC_STARVING ) {
+		++lc->starving;
+		RETURN(NULL);
+	}
+
+	e = lc_find(lc, enr);
+	if (e) {
+		++lc->hits;
+		if( e->refcnt++ == 0) lc->used++;
+		list_move(&e->list,&lc->in_use); // Not evictable...
+	}
+	RETURN(e);
+}
+
 void lc_changed(struct lru_cache* lc, struct lc_element* e)
 {
 	PARANOIA_ENTRY();
 	BUG_ON(e != lc->changing_element);
+	++lc->changed;
 	e->lc_number = lc->new_number;
 	list_add(&e->list,&lc->in_use);
 	hlist_add_head( &e->colision, lc->slot + lc_hash_fn(lc, lc->new_number) );
@@ -257,8 +310,10 @@ unsigned int lc_put(struct lru_cache* lc, struct lc_element* e)
 
 	PARANOIA_ENTRY();
 	BUG_ON(e->refcnt == 0);
+	BUG_ON(e == lc->changing_element);
 	if ( --e->refcnt == 0) {
 		list_move(&e->list,&lc->lru); // move it to the front of LRU.
+		lc->used--;
 		clear_bit(__LC_STARVING,&lc->flags);
 		smp_mb__after_clear_bit();
 	}
@@ -287,3 +342,29 @@ void lc_set(struct lru_cache* lc, unsigned int enr, int index)
 	list_move(&e->list, e->refcnt ? &lc->in_use : &lc->lru);
 }
 
+#if 0
+/**
+ * lc_dump: Dump a complete LRU cache to seq in textual form.
+ */
+void lc_dump(struct lru_cache* lc, struct seq_file *seq, char* utext,
+	     void (*detail) (struct seq_file *, struct lc_element *) )
+{
+	unsigned int nr_elements = lc->nr_elements;
+	struct lc_element *e;
+	int i;
+
+	seq_printf(seq,"\tnn: lc_number refcnt %s\n ",utext);
+	for(i=0;i<nr_elements;i++) {
+		e = lc_entry(lc,i);
+		if( e->lc_number == LC_FREE ) {
+			seq_printf(seq,"\t%2d: FREE\n",i );
+		} else {
+			seq_printf(seq,"\t%2d: %4u %4u    ", i,
+				   e->lc_number,
+				   e->refcnt );
+			detail(seq,e);
+		}
+	}
+}
+
+#endif
