@@ -1,12 +1,13 @@
 /*
 -*- linux-c -*-
    lru_cache.c
-   Kernel module for 2.4.x/2.6.x Kernels
+   Kernel module for 2.6.x Kernels
 
-   This file is part of drbd by Philipp Reisner.
+   This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2003-2004, Philipp Reisner <philipp.reisner@linbit.com>.
-        main author.
+   Copyright (C) 2003-2007, LINBIT Information Technologies GmbH.
+   Copyright (C) 2003-2007, Philipp Reisner <philipp.reisner@linbit.com>.
+   Copyright (C) 2003-2007, Lars Ellenberg <lars.ellenberg@linbit.com>.
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -28,7 +29,7 @@
   The lru_cache describes a big set of objects that are addressed
   by an index number (=lc_number). Only a small fraction of this set
   is present in the cache.
-  (You set the size of the cache using lc_resize)
+  (You set the size of the cache during lc_alloc)
   Once created, the api consists of
     lc_find(,nr) -- finds the object with the given number, if present
     lc_get(,nr)  -- finds the object and increases the usage count
@@ -52,26 +53,10 @@
 #endif
 
 #include <linux/version.h>
-#if LINUX_VERSION_CODE < KERNEL_VERSION (2,4,20)
-static inline void list_move(struct list_head *list, struct list_head *head)
-{
-        __list_del(list->prev, list->next);
-        list_add(list, head);
-}
-#endif
 
-#ifndef max
-// For RH 2.4.9
-# define max(x,y) \
-	({ typeof(x) __x = (x); typeof(y) __y = (y); \
-	   (void)(&__x == &__y); \
-	   __x > __y ? __x: __y; })
-#endif
-
-#ifndef BUG_ON
-	/* for ancient 2.4 kernels */
-# define BUG_ON(condition) do { if (unlikely((condition)!=0)) BUG(); } while(0)
-#endif
+/* FIXME
+ * I want these structs opaque outside of lru_cache.c
+ */
 
 struct lc_element {
 	struct hlist_node colision;
@@ -87,10 +72,16 @@ struct lru_cache {
 	size_t element_size;
 	unsigned int  nr_elements;
 	unsigned int  new_number;
+
+	/* here may or may not be a pad... */
+
+	unsigned int used;
 	unsigned long flags;
+	unsigned long hits, misses, starving, dirty, changed;
 	struct lc_element *changing_element; // just for paranoia
 
 	void  *lc_private;
+	const char *name;
 
 	struct hlist_head slot[0];
 	// hash colision chains here, then element storage.
@@ -107,17 +98,23 @@ enum {
 #define LC_DIRTY    (1<<__LC_DIRTY)
 #define LC_STARVING (1<<__LC_STARVING)
 
-extern struct lru_cache* lc_alloc(unsigned int e_count, size_t e_size,
-				  void *private_p);
+extern struct lru_cache* lc_alloc(const char *name, unsigned int e_count,
+				  size_t e_size, void *private_p);
 extern void lc_free(struct lru_cache* lc);
 extern void lc_set (struct lru_cache* lc, unsigned int enr, int index);
 extern void lc_del (struct lru_cache* lc, struct lc_element *element);
 
+extern struct lc_element* lc_try_get(struct lru_cache* lc, unsigned int enr);
 extern struct lc_element* lc_find(struct lru_cache* lc, unsigned int enr);
 extern struct lc_element* lc_get (struct lru_cache* lc, unsigned int enr);
 extern unsigned int       lc_put (struct lru_cache* lc, struct lc_element* e);
 extern void            lc_changed(struct lru_cache* lc, struct lc_element* e);
 
+struct seq_file;
+extern size_t lc_printf_stats(struct seq_file *seq, struct lru_cache* lc);
+
+void lc_dump(struct lru_cache* lc, struct seq_file *seq, char* utext,
+	     void (*detail) (struct seq_file *, struct lc_element *) );
 
 /* This can be used to stop lc_get from changing the set of active elements.
  * Note that the reference counts and order on the lru list may still change.
@@ -134,7 +131,13 @@ static inline void lc_unlock(struct lru_cache* lc)
 	smp_mb__after_clear_bit();
 }
 
-#define LC_FREE (-1)
+static inline int lc_is_used(struct lru_cache* lc, unsigned int enr)
+{
+	struct lc_element* e = lc_find(lc,enr);
+	return (e && e->refcnt);
+}
+
+#define LC_FREE (-1U)
 
 #define lc_e_base(lc)  ((char*) ( (lc)->slot + (lc)->nr_elements ) )
 #define lc_entry(lc,i) ((struct lc_element*) \
