@@ -3,9 +3,9 @@
 
    This file is part of DRBD by Philipp Reisner and Lars Ellenberg.
 
-   Copyright (C) 2006-2008, LINBIT Information Technologies GmbH
-   Copyright (C) 2006-2008, Philipp Reisner <philipp.reisner@linbit.com>
-   Copyright (C) 2006-2008, Lars Ellenberg  <lars.ellenberg@linbit.com>
+   Copyright (C) 2006-2007, LINBIT Information Technologies GmbH
+   Copyright (C) 2006-2007, Philipp Reisner <philipp.reisner@linbit.com>
+   Copyright (C) 2006-2007, Lars Ellenberg  <lars.ellenberg@linbit.com>
 
    drbd is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,7 +36,6 @@ enum range_checks
 	R_MAX_BUFFERS,
 	R_MAX_EPOCH_SIZE,
 	R_SNDBUF_SIZE,
-	R_RCVBUF_SIZE,
 	R_KO_COUNT,
 	R_RATE,
 	R_GROUP,
@@ -45,15 +44,12 @@ enum range_checks
 	R_META_IDX,
 	R_WFC_TIMEOUT,
 	R_DEGR_WFC_TIMEOUT,
-	R_OUTDATED_WFC_TIMEOUT,
 };
 
 enum yytokentype {
 	TK_GLOBAL = 258,
 	TK_RESOURCE,
 	TK_ON,
-	TK_STACKED,
-	TK_IGNORE,
 	TK_NET,
 	TK_DISK,
 	TK_SKIP,
@@ -62,18 +58,19 @@ enum yytokentype {
 	TK_DISABLE_IP_VERIFICATION,
 	TK_DIALOG_REFRESH,
 	TK_PROTOCOL,
-	TK_HANDLER,
+	TK_HANDLER, // 270
 	TK_COMMON,
 	TK_ADDRESS,
 	TK_DEVICE,
-	TK_MINOR,
 	TK_META_DISK,
 	TK_FLEX_META_DISK,
 	TK_MINOR_COUNT,
+
 	TK_IPADDR,
 	TK_INTEGER,
 	TK_STRING,
-	TK_ELSE,
+	TK_ELSE, // 280
+
 	TK_DISK_SWITCH,
 	TK_DISK_OPTION,
 	TK_NET_SWITCH,
@@ -82,33 +79,16 @@ enum yytokentype {
 	TK_SYNCER_OPTION,
 	TK_STARTUP_SWITCH,
 	TK_STARTUP_OPTION,
-	TK_STARTUP_DELEGATE,
 	TK_HANDLER_OPTION,
-	TK_USAGE_COUNT,
+	TK_USAGE_COUNT,  // 290
 	TK_ASK,
 	TK_YES,
 	TK_NO,
 	TK__IS_DEFAULT,
+	TK__MAJOR,
+	TK__MINOR,
 	TK__THIS_HOST,
-	TK__REMOTE_HOST,
-	TK_PROXY,
-	TK_INSIDE,
-	TK_OUTSIDE,
-	TK_MEMLIMIT,
-	TK_PROXY_OPTION,
-	TK_PROXY_SWITCH,
-	TK_ERR_STRING_TOO_LONG,
-	TK_ERR_DQSTRING_TOO_LONG,
-	TK_ERR_DQSTRING,
-	TK_SCI,
-	TK_SDP,
-	TK_SSOCKS,
-	TK_IPV4,
-	TK_IPV6,
-	TK_IPADDR6,
-	TK_NET_DELEGATE,
-	TK_INCLUDE,
-	TK_FLOATING,
+	TK__REMOTE_HOST
 };
 
 typedef struct YYSTYPE {
@@ -125,5 +105,3 @@ extern char* yytext;
 
 /* avoid compiler warnings about implicit declaration */
 int yylex(void);
-void my_yypush_buffer_state(FILE *f);
-void yypop_buffer_state (void );
